/*
    Copyright (C) 2005-2007  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <glib.h>
#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#include <wspiapi.h>
#else
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif
#include <openssl/ssl.h>

#include "strptime.h"
#include "nntpgrab_plugin_nntp.h"

static PluginNNTPImportedFuncs imported_funcs;

// Keep track of the number of bytes received for the last 10 seconds
static int bytes_received[10];
static time_t timestamp = 0;
static GThread *traffic_thread = NULL;
static gboolean abort_traffic_thread = FALSE;
static GStaticRWLock traffic_rwlock = G_STATIC_RW_LOCK_INIT;

// Link a socket_id to a SSL pointer
static GHashTable *connections = NULL;
static GStaticRWLock connections_rwlock = G_STATIC_RW_LOCK_INIT;

typedef struct _connection_data {
    int socket_id;
    gboolean use_ssl;
    SSL *ssl;
} ConnectionData;

#ifdef WIN32
#define CLOSE(x)    closesocket(x); remove_connection(x)
#else
#define CLOSE(x)    close(x); remove_connection(x)
#endif

gpointer
traffic_thread_func(gpointer data)
{
    while (!abort_traffic_thread) {
        time_t prev_timestamp;

        // Traffic monitoring
        prev_timestamp = timestamp;
        timestamp = time(NULL);

        if (timestamp > prev_timestamp) {
            time_t diff;

            // Calculate how many steps we need to shift
            diff = timestamp - prev_timestamp;
            if (diff > 10) {
                diff = 10;
            }

            g_static_rw_lock_writer_lock(&traffic_rwlock);

            // Notify upstream about our values
            imported_funcs.update_traffic_monitor(bytes_received, timestamp);

            // Shift all the previous values by 'diff'
            memmove(bytes_received, &bytes_received[diff], (10 - diff) * sizeof(int));
            memset(&bytes_received[10 - diff], 0, (10 - diff) * sizeof(int));

            g_static_rw_lock_writer_unlock(&traffic_rwlock);

#if 0
            g_print("stamp = %li\n", timestamp);
            for (i = 0; i < 10; i++) {
                g_print("bytes_received[%i] = %i\n", i, bytes_received[i]);
            }
            g_print("\n");
#endif
        }

        g_usleep(G_USEC_PER_SEC * 1);
    }

    return NULL;
}

gboolean
nntpgrab_plugin_nntp_initialize(PluginNNTPImportedFuncs funcs)
{
    imported_funcs = funcs;

    memset(bytes_received, 0, sizeof(bytes_received));
    abort_traffic_thread = FALSE;
    traffic_thread = g_thread_create(traffic_thread_func, NULL, TRUE, NULL);
    connections = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, g_free);

    return TRUE;
}

int
nntpgrab_plugin_nntp_get_version(void)
{
    return NNTPGRAB_PLUGIN_API_VERSION;
}

gboolean
nntpgrab_plugin_nntp_get_is_connected(int conn_id)
{

    if (conn_id <= 0) {
        return FALSE;
    } else {
        return TRUE;
    }
}

static int
get_status_code(const char *line)
{
    /* The status code is mentioned at the first 3 characters of the line.
     * atoi() doesn't detect errors and quits parsing after the first invalid
     * character which is enough for us */
    if (!line) {
        return -1;
    }

    return atoi(line);
}

static void
perform_traffic_monitoring(int bytes_read, gpointer data)
{
    g_static_rw_lock_writer_lock(&traffic_rwlock);
    bytes_received[9] += bytes_read;
    g_static_rw_lock_writer_unlock(&traffic_rwlock);
}

static void
add_connection(int socket_id, gboolean use_ssl, SSL *ssl)
{
    ConnectionData *data = g_malloc0(sizeof(ConnectionData));

    data->socket_id = socket_id;
    data->use_ssl = use_ssl;
    data->ssl = ssl;

    g_static_rw_lock_writer_lock(&connections_rwlock);
    g_hash_table_replace(connections, GINT_TO_POINTER(socket_id), data);
    g_static_rw_lock_writer_unlock(&connections_rwlock);
}

static void
remove_connection(int socket_id)
{
    ConnectionData *data;

    g_static_rw_lock_reader_lock(&connections_rwlock);
    data = g_hash_table_lookup(connections, GINT_TO_POINTER(socket_id));
    g_static_rw_lock_reader_unlock(&connections_rwlock);

    if (!data) {
        return;
    }

    if (data->ssl) {
        SSL_free(data->ssl);
    }

    g_static_rw_lock_writer_lock(&connections_rwlock);
    g_hash_table_remove(connections, GINT_TO_POINTER(socket_id));
    g_static_rw_lock_writer_unlock(&connections_rwlock);
}

static int
perform_peek(int socket_id, void *buf, int len)
{
    ConnectionData *data;

    g_static_rw_lock_reader_lock(&connections_rwlock);
    data = g_hash_table_lookup(connections, GINT_TO_POINTER(socket_id));
    g_static_rw_lock_reader_unlock(&connections_rwlock);

    if (!data) {
        errno = EBADF;
        return -1;
    }

    if (data->use_ssl) {
        return SSL_peek(data->ssl, buf, len);
    } else {
        return recv(socket_id, buf, len, MSG_PEEK);
    }
}

static int
perform_read(int socket_id, void *buf, int len)
{
    ConnectionData *data;

    g_static_rw_lock_reader_lock(&connections_rwlock);
    data = g_hash_table_lookup(connections, GINT_TO_POINTER(socket_id));
    g_static_rw_lock_reader_unlock(&connections_rwlock);

    if (!data) {
        errno = EBADF;
        return -1;
    }

    if (data->use_ssl) {
        return SSL_read(data->ssl, buf, len);
    } else {
        return recv(socket_id, buf, len, 0);
    }
}

static int
perform_write(int socket_id, const void *buf, int len)
{
    ConnectionData *data;

    g_static_rw_lock_reader_lock(&connections_rwlock);
    data = g_hash_table_lookup(connections, GINT_TO_POINTER(socket_id));
    g_static_rw_lock_reader_unlock(&connections_rwlock);

    if (!data) {
        errno = EBADF;
        return -1;
    }

    if (data->use_ssl) {
        return SSL_write(data->ssl, buf, len);
    } else {
        return send(socket_id, buf, len, 0);
    }
}

static char *
read_line(int socket_id, gpointer data)
{
    char buf[4096];      // It should be safe to assume that one line is less than 4096 characters
    char *newline_pos;
    char *line;
    int i;
    int bytes_read;
    time_t stamp;

    memset(buf, 0, sizeof(buf));

    stamp = time(NULL);

    do {
        if (perform_peek(socket_id, buf, sizeof(buf)) <= 0 && errno != 0) {
            // Did we receive a interrupt signal?
            if (errno == EINTR) {
                // Ignore, try again
                g_print(__FILE__":%i EINTR received while performing MSG_PEEK\n", __LINE__);
                continue;
            }

            g_print(__FILE__":%i unknown error received while performing MSG_PEEK: errno = %i, %s\n", __LINE__, errno, strerror(errno));
            
            // Read error, disconnect
            imported_funcs.notify_disconnect(DISCONNECT_READ_ERROR, strerror(errno), data);

            CLOSE(socket_id);

            return NULL;
        }

        // Is there a newline character in buf ?
        newline_pos = NULL;
        for (i = 0; i < sizeof(buf); i++) {
            if (buf[i] == '\n') {
                newline_pos = &buf[i];
                break;
            }
        }

        if (newline_pos) {
            break;
        }

        g_usleep(1000);

        if (time(NULL) > stamp + 30) {
            // Read timeout
            imported_funcs.notify_disconnect(DISCONNECT_READ_TIMEOUT, strerror(errno), data);

            CLOSE(socket_id);

            g_print(__FILE__":%i read timeout: errno = %i, %s\n", __LINE__, errno, strerror(errno));

            return NULL;
        }
    } while (TRUE);

    // We know where the end of the line is, read enough characters from the socket
    line = g_malloc0(newline_pos - buf + 2);
    if ((bytes_read = perform_read(socket_id, line, newline_pos - buf + 1)) == -1) {
        g_print("line = %s\n", line);
        g_print(__FILE__":%i unknown error received while performing recv: errno = %i, %s\n", __LINE__, errno, strerror(errno));
        imported_funcs.notify_disconnect(DISCONNECT_READ_ERROR, strerror(errno), data);

        g_free(line);

        CLOSE(socket_id);

        return NULL;
    }

    perform_traffic_monitoring(bytes_read, data);

    return line;
}

static gboolean
write_line(int socket_id, const char *line, gpointer data)
{
    if (perform_write(socket_id, line, strlen(line)) == -1) {
        imported_funcs.notify_disconnect(DISCONNECT_WRITE_ERROR, strerror(errno), data);

        CLOSE(socket_id);

        return FALSE;
    }

    return TRUE;
}

static NNTPGrabErrCode
login(const char *username, const char *password, char **errmsg, int *conn_id, gpointer data)
{
    char *line;
    char *welcome_msg;

    welcome_msg = read_line(*conn_id, data);
    if (!welcome_msg) return NNTP_ERROR_READ;

    // Strip the newline character
    if (welcome_msg[strlen(welcome_msg) - 1]  == '\n') {
        welcome_msg[strlen(welcome_msg) - 1] = '\0';
    }

    if (welcome_msg[strlen(welcome_msg) - 1]  == '\r') {
        welcome_msg[strlen(welcome_msg) - 1] = '\0';
    }

    switch(get_status_code(welcome_msg)) {
        case 200:
        case 201:
            // Login OK
            break;

        default:
            // Login not OK..probably too many connections from this address
            if (errmsg) {
                *errmsg = g_strdup(welcome_msg);
            }
            g_free(welcome_msg);
            CLOSE(*conn_id);

            return NNTP_ERROR_TOO_MANY_CONNECTIONS;
    }

    // Do we need to login using a username and password ?
    if (username && strlen(username) > 0 && password && strlen(password) > 0) {
        line = g_strdup_printf("AUTHINFO USER %s\r\n", username);
        if (!write_line(*conn_id, line, data)) {
            if (errmsg) {
                line[strlen(line) - 1] = '\0';
                *errmsg = g_strdup(line);
            }
            g_free(line);
            return NNTP_ERROR_WRITE;
        }

        g_free(line);

        line = read_line(*conn_id, data);
        if (!line) return NNTP_ERROR_READ;

        if (get_status_code(line) != 381) {
            // Invalid reply, disconnect
            if (errmsg) {
                *errmsg = g_strdup(line);
            }
            g_free(line);
            CLOSE(*conn_id);

            return NNTP_ERROR_INVALID_MSG;
        }

        g_free(line);

        line = g_strdup_printf("AUTHINFO PASS %s\r\n", password);
        if (!write_line(*conn_id, line, data)) {
            if (errmsg) {
                line[strlen(line) - 1] = '\0';
                *errmsg = g_strdup(line);
            }
            g_free(line);
            return NNTP_ERROR_WRITE;
        }

        g_free(line);

        line = read_line(*conn_id, data);
        if (!line) return NNTP_ERROR_READ;

        if (get_status_code(line) != 281) {
            // Invalid reply, disconnect
            if (errmsg) {
                *errmsg = g_strdup(line);
            }
            g_free(line);
            CLOSE(*conn_id);

            return NNTP_ERROR_INVALID_MSG;
        }

        g_free(line);
    }

    // Tell the usenet server we want to use the READER mode
    line = g_strdup("MODE READER\r\n");
    if (!write_line(*conn_id, line, data)) {
        if (errmsg) {
            line[strlen(line) - 1] = '\0';
            *errmsg = g_strdup(line);
        }
        g_free(line);

        CLOSE(*conn_id);

        return NNTP_ERROR_WRITE;
    }

    g_free(line);

    line = read_line(*conn_id, data);
    if (!line) return NNTP_ERROR_READ;

    switch(get_status_code(line)) {
        case 200:
        case 201:
            g_free(line);
            break;

        case 480:
            // Some usenet servers use this to notify there are too many connections active
            if (errmsg) {
                *errmsg = g_strdup(line);
            }
            g_free(line);
            CLOSE(*conn_id);

            return NNTP_ERROR_TOO_MANY_CONNECTIONS;

        default:
            // Invalid reply, disconnect
            if (errmsg) {
                *errmsg = g_strdup(line);
            }
            g_free(line);
            CLOSE(*conn_id);

            return NNTP_ERROR_INVALID_MSG;
    }

    // Login complete, let's party!

    imported_funcs.notify_connected(welcome_msg, data);
    g_free(welcome_msg);

    return NNTP_ERROR_NONE;
}

static SSL *
prepare_ssl_connection(int conn_id, char **errmsg)
{
    SSL_CTX     *ctx;
    SSL         *ssl;
    SSL_METHOD  *meth;
    int          err;
#if 0
    char        *str;
    X509        *server_cert;
#endif

    SSLeay_add_ssl_algorithms();
    meth = SSLv2_client_method();
    SSL_load_error_strings();
    ctx = SSL_CTX_new (meth);

    if (!ctx) {
        if (errmsg) {
            *errmsg = g_strdup("SSL_CTX_new FAILED");
        }

        return NULL;
    }

    ssl = SSL_new (ctx);
    SSL_CTX_free(ctx);

    if (!ssl) {
        if (errmsg) {
            *errmsg = g_strdup("SSL_new FAILED");
        }

        return NULL;
    }

    SSL_set_fd (ssl, conn_id);
    err = SSL_connect (ssl);
    if (!err) {
        if (errmsg) {
            *errmsg = g_strdup("SSL_connect FAILED");
        }

        return NULL;
    }

#if 0
    printf ("SSL connection using %s\n", SSL_get_cipher (ssl));
    server_cert = SSL_get_peer_certificate (ssl);
    if (server_cert) {
        printf ("Server certificate:\n");

        str = X509_NAME_oneline (X509_get_subject_name (server_cert),0,0);
        if (str) {
            printf ("\t subject: %s\n", str);
            OPENSSL_free (str);
        }

        str = X509_NAME_oneline (X509_get_issuer_name  (server_cert),0,0);
        if (str) {
            printf ("\t issuer: %s\n", str);
            OPENSSL_free (str);
        }

        /* We could do all sorts of certificate verification stuff here before
           deallocating the certificate. */

        X509_free (server_cert);
    }
#endif

    return ssl;
}

NNTPGrabErrCode
nntpgrab_plugin_nntp_connect(const char *hostname, int port, const char *username, const char *password, gboolean use_ssl, char **errmsg, int *conn_id, gpointer data)
{
    struct addrinfo hints, *res, *ressave;
    int n;
    char *str_port;
    int last_errno;
    struct timeval tv;
    SSL *ssl;

    g_assert(hostname);
    g_assert(port > 0);
    g_assert(port < 65535);

    *conn_id = -1;

    memset(&hints, 0, sizeof(struct addrinfo));

    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    str_port = g_strdup_printf("%i", port);
    n = getaddrinfo(hostname, str_port, &hints, &res);
    g_free(str_port);

    if (n != 0) {
        if (errmsg) {
            *errmsg = g_strdup(gai_strerror(n));
        }
        return NNTP_ERROR_HOST_NOT_FOUND;
    }

    ressave = res;
    last_errno = 0;

    while (res) {
        *conn_id = socket(res->ai_family, res->ai_socktype, res->ai_protocol);

#ifdef WIN32
        if ((*conn_id == INVALID_SOCKET)) {
#else
        if ((*conn_id == -1)) {
#endif
            // The socket couldn't be created
            // Save the errno and try the next item in the list
            last_errno = errno;
            res = res->ai_next;
            *conn_id = -1;

            if (errmsg) {
                if (*errmsg) {
                    g_free(*errmsg);
                }

                *errmsg = g_strdup(strerror(last_errno));
            }

            continue;
        }

        // Set the connection timeout on the socket
        tv.tv_sec = 5;
        tv.tv_usec = 0;
        setsockopt(*conn_id, SOL_SOCKET, SO_RCVTIMEO, (const void*) &tv, sizeof(tv));
        setsockopt(*conn_id, SOL_SOCKET, SO_SNDTIMEO, (const void*) &tv, sizeof(tv));

        // Try to connect to the server
        imported_funcs.notify_connecting(*conn_id, data);

        if (connect(*conn_id, res->ai_addr, (int) res->ai_addrlen) == 0) {
            // Connection succesfull
            last_errno = 0;
            break;
        }

        // Connection could not be made, save the errno
#ifndef WIN32
        if (errno == EINPROGRESS) {
            // The connection could not be established within the connect time limit.
            // This provides a more clear error message
            last_errno = ETIMEDOUT;
        } else
#endif
        {
            last_errno = errno;
        }

        if (errmsg) {
            if (*errmsg) {
                g_free(*errmsg);
            }

            *errmsg = g_strdup(strerror(last_errno));
        }

#if WIN32
        if (last_errno == WSAETIMEDOUT) {
#else
        if (last_errno == ETIMEDOUT) {
#endif
            imported_funcs.notify_disconnect(DISCONNECT_CONNECT_TIMEOUT, (errmsg) ? *errmsg : NULL, data);
        } else {
            imported_funcs.notify_disconnect(DISCONNECT_CONNECTION_REFUSED, (errmsg) ? *errmsg : NULL, data);
        }

        CLOSE(*conn_id);
        *conn_id = -1;

        res = res->ai_next;
    }

    if (*conn_id == -1) {
        freeaddrinfo(ressave);

        switch (last_errno) {
#if WIN32
            case WSAETIMEDOUT:
#else
            case ETIMEDOUT:
#endif
                return NNTP_ERROR_CONNECTION_TIMEOUT;

            default:
                return NNTP_ERROR_CONNECTION_REFUSED;
        };
    }

    freeaddrinfo(ressave);

    if (use_ssl) {
        if ((ssl = prepare_ssl_connection(*conn_id, errmsg)) == NULL) {
            CLOSE(*conn_id);
            return NNTP_ERROR_SSL_INITIALISE;
        }
    } else {
        ssl = NULL;
    }

    add_connection(*conn_id, use_ssl, ssl);

    return login(username, password, errmsg, conn_id, data);
}

void
nntpgrab_plugin_nntp_disconnect(int conn_id, gpointer data)
{
    CLOSE(conn_id);

    imported_funcs.notify_disconnect(DISCONNECT_NORMAL, NULL, data);
}

struct _list_elem
{
    char data[65535];
    int length;
};

NNTPGrabErrCode
nntpgrab_plugin_nntp_get_article_by_message_id(int conn_id, const char *message_id, const char **headers, void **data, int *data_len, gpointer cb_data)
{
    char *line;
    GList *list_data = NULL;
    GList *list;
    int total_length = 0;
    char *newline_pos;
    char *ptr;
    int prev_progress_update = -1;

    if (headers) {
        g_error(__FILE__":%i fetching of headers not yet implemented!\n", __LINE__);
    }

    line = g_strdup_printf("BODY %s\r\n", message_id);
    if (!write_line(conn_id, line, cb_data)) {
        g_free(line);
        return NNTP_ERROR_WRITE;
    }

    g_free(line);

    line = read_line(conn_id, cb_data);
    if (!line) return NNTP_ERROR_READ;

    imported_funcs.notify_download_begin(cb_data);

    switch (get_status_code(line)) {
        case 222:            // article retrieved - body follows
            break;

        case 430:            // no such article found
            g_free(line);

            imported_funcs.notify_part_not_available(cb_data);

            return NNTP_ERROR_PART_NOT_AVAILABLE;

        case 220:            // article retrieved - head and body follow
        case 221:            // article retrieved - head follows
        case 223:            // article retrieved - request text separately
        case 412:            // no newsgroup has been selected
        case 420:            // no current article has been selected
        case 423:            // no such article number in this group
        default:
            imported_funcs.notify_disconnect(DISCONNECT_INVALID_MSG, line, cb_data);
            g_free(line);
            CLOSE(conn_id);

            return NNTP_ERROR_INVALID_MSG;
    }

    g_free(line);

    while (TRUE) {
        struct _list_elem *elem;
        char *end_of_data;
        fd_set rfds;
        struct timeval tv;
        int retval;

        elem = g_malloc0(sizeof(struct _list_elem));

        // Guard for read timeouts
        FD_ZERO(&rfds);
        FD_SET(conn_id, &rfds);

        /* Wait up to five seconds. */
        tv.tv_sec = 30;
        tv.tv_usec = 0;

        retval = select(conn_id + 1, &rfds, NULL, NULL, &tv);
        if (retval <= 0) {
            // Read timeout
            imported_funcs.notify_disconnect(DISCONNECT_READ_TIMEOUT, strerror(errno), cb_data);

            CLOSE(conn_id);

            list = list_data;
            while (list) {
                g_free(list->data);
                list = list->next;
            }

            g_list_free(list_data);

            return NNTP_ERROR_READ;
        }

        // There should be data waiting now
        elem->length = perform_read(conn_id, elem->data, sizeof(elem->data));
        if (elem->length == -1) {
            imported_funcs.notify_disconnect(DISCONNECT_READ_ERROR, strerror(errno), cb_data);

            CLOSE(conn_id);

            list = list_data;
            while (list) {
                g_free(list->data);
                list = list->next;
            }

            g_list_free(list_data);

            return NNTP_ERROR_READ;
        }

        perform_traffic_monitoring(elem->length, cb_data);

        // Is the chunk large enough to detect the end of data ?
        if (elem->length < 4) {
            // Nope, add the data to the previous chunk
            GList *last_item;

            last_item = g_list_last(list_data);
            if (!last_item) {
                // List is still empty, just add the element to the list
                list_data = g_list_append(list_data, elem);
                total_length += elem->length;
                continue;
            }

            struct _list_elem *prev_elem;
            char *tmp;

            prev_elem = last_item->data;

            tmp = g_malloc0(prev_elem->length + elem->length);
            memcpy(tmp, prev_elem->data, prev_elem->length);
            memcpy(tmp + prev_elem->length, elem->data, elem->length);

            memcpy(prev_elem->data, tmp, prev_elem->length + elem->length);
            prev_elem->length += elem->length;

            g_free(tmp);

            g_free(elem);
            elem = prev_elem;
        } else {
            // Chunk is large enough, add it to the list
            list_data = g_list_append(list_data, elem);
        }

        // Check if we are at the end of the message
        end_of_data = elem->data + elem->length - 4;
        if (!strcmp(end_of_data, "\n.\r\n") || !strcmp(end_of_data + 1, "\n.\n")) {
            // Only keep the first part of the chunk
            if (!strcmp(end_of_data, "\n.\r\n")) {
                elem->length -= 3;
            } else {
                elem->length -= 2;
            }

            total_length += elem->length;

            // Break out of the loop as we are at the end of the message
            break;
        }

        total_length += elem->length;

        // Send a progress update every 10KB
        if ((total_length / (1024 * 10)) > prev_progress_update) {
            prev_progress_update = total_length / (1024 * 10);
            imported_funcs.update_progress(total_length, cb_data);
        }
    }

    imported_funcs.update_progress(total_length, cb_data);

    // All the data is now received. Glue it together in one big chunk of data
    ptr = g_malloc(total_length + 1);
    *data_len = 0;
    *data = ptr;

    list = list_data;

    while (list) {
        struct _list_elem *elem = list->data;
        int pos = *data_len;

        g_assert(pos < total_length);

        memcpy(ptr + pos, elem->data, elem->length);
        *data_len += elem->length;

        g_free(elem);

        list = list->next;
    }

    g_list_free(list_data);

    // Do we need to un-escape any double-dots ?
    newline_pos = strstr(ptr, "\n");
    while (newline_pos) {
        if (newline_pos[0] == '\n' && newline_pos[1] == '.' && newline_pos[2] == '.') {
            char *begin = ptr;
            memmove(newline_pos + 1, newline_pos + 2, *data_len - (newline_pos - begin) - 1);
            *data_len -= 1;
        }
        newline_pos = strstr(newline_pos + 1, "\n");
    }

    imported_funcs.notify_download_complete(cb_data);

    return NNTP_ERROR_NONE;
}

NNTPGrabErrCode
nntpgrab_plugin_nntp_group(int conn_id, const char *newsgroup, int *num_articles, int *start_article, int *end_article, gpointer data)
{
    char *line;

    g_assert(newsgroup);
    g_assert(num_articles);
    g_assert(start_article);
    g_assert(end_article);

    line = g_strdup_printf("GROUP %s\r\n", newsgroup);
    if (!write_line(conn_id, line, data)) {
        g_free(line);
        return NNTP_ERROR_WRITE;
    }

    line = read_line(conn_id, data);
    if (!line) return NNTP_ERROR_READ;

    switch (get_status_code(line)) {
        case 211:
            /*  211 n f l s group selected
                   (n = estimated number of articles in group,
                    f = first article number in the group,
                    l = last article number in the group,
                    s = name of the group.)
            */
            sscanf(line, "%*i %i %i %i %*s", num_articles, start_article, end_article);
            g_free(line);
            return NNTP_ERROR_NONE;

        case 411:
            // No such group
            g_free(line);
            return NNTP_ERROR_NO_SUCH_GROUP;

        default:
            // Invalid response
            imported_funcs.notify_disconnect(DISCONNECT_INVALID_MSG, line, data);
            g_free(line);
            CLOSE(conn_id);

            return NNTP_ERROR_INVALID_MSG;
    }
}

NNTPGrabErrCode
nntpgrab_plugin_nntp_xover(int conn_id, int start_range, int end_range, gpointer data)
{
    char *line;
    char **parts;
    struct tm tm;
    time_t post_date;

    g_assert(start_range >= 1);
    g_assert(end_range >= -1);

    if (end_range == -1) {
        line = g_strdup_printf("XOVER %i-\r\n", start_range);
    } else {
        line = g_strdup_printf("XOVER %i-%i\r\n", start_range, end_range);
    }

    if (!write_line(conn_id, line, data)) {
        g_free(line);
        return NNTP_ERROR_WRITE;
    }

    g_free(line);

    line = read_line(conn_id, data);
    if (!line) return NNTP_ERROR_READ;

    switch (get_status_code(line)) {
        case 224:
            // 224 Overview information follows
            g_free(line);
            break;

        default:
            // Invalid response
            imported_funcs.notify_disconnect(DISCONNECT_INVALID_MSG, line, data);
            g_free(line);
            CLOSE(conn_id);

            return NNTP_ERROR_INVALID_MSG;
    }

    while (1) {
        line = read_line(conn_id, data);
        if (!line) return NNTP_ERROR_READ;

        if (strncmp(line, ".", 1) == 0) {
            g_free(line);
            break;
        }

        parts = g_strsplit(line, "\t", 0);

        /* Prevent segfaults */
        if (!parts[0]) {
            g_strfreev(parts);
            g_free(line);
            continue;
        } else if (!parts[1]) {
            g_strfreev(parts);
            g_free(line);
            continue;
        } else if (!parts[2]) {
            g_strfreev(parts);
            g_free(line);
            continue;
        } else if (!parts[3]) {
            g_strfreev(parts);
            g_free(line);
            continue;
        } else if (!parts[4]) {
            g_strfreev(parts);
            g_free(line);
            continue;
        } else if (!parts[5]) {
            g_strfreev(parts);
            g_free(line);
            continue;
        } else if (!parts[6]) {
            g_strfreev(parts);
            g_free(line);
            continue;
        } else if (!parts[7]) {
            g_strfreev(parts);
            g_free(line);
            continue;
        }

        memset(&tm, 0, sizeof(tm));
        if (strptime(parts[3], "%d %b %Y %H:%M:%S %Z", &tm)) {
            post_date = mktime(&tm);
        } else {
            memset(&tm, 0, sizeof(tm));
            if (strptime(parts[3], "%A, %d %b %Y %H:%M:%S %Z", &tm)) {
                post_date = mktime(&tm);
            } else {
                // Date could not be parsed
                post_date = 0;

                g_strfreev(parts);
                g_free(line);
                continue;
            }
        }

        imported_funcs.parse_header(atoi(parts[0]), parts[1], parts[2], post_date, parts[4], atoi(parts[6]), atoi(parts[7]), start_range, end_range, data);

        g_strfreev(parts);
        g_free(line);
    }

    return TRUE;
}
